(* Computes Hecke matrices in raw format, i.e. with respect to different input
   and output bases, and prepares pari script to extract consistent description
   of eigenvalues. Unlike pheckeE it expects degree four extension.
*)
program hecke ( input,output,lnkone,lnktwo )

; const
    maxd	= 500       (* changed by Bert, 14-9-93, was 200 *)
    (* maxd	= 200 *)
  ; maxorbits	= 6000
  ; maxparams	= 6000
  ; maxdim	= 200     (* changed 4-9-93 by Bert *)

; type
    sign	= -1..1
  ; vector3	= array [1..3] of integer
  ; matrix3x3	= array[1..3,1..3] of integer
  ; sgdvector3	= record x : vector3 ; sg : sign end

; var
    lnkone,lnktwo    :text
  ; d,orbits,params,dim,ell	: integer
  ; gcdwithd,invmodd	: array [1..maxd] of integer
  ; orbit		: array [1..maxorbits] of
			  record
			    x : vector3
			  ; orbref,parref : integer
			  end
  ; param		: array [1..maxparams] of integer
  ; soln		: array [1..maxdim,1..maxparams] of integer
  ; freepar		: array [1..maxdim] of integer
  ; raw			: array [1..maxdim,1..maxparams] of integer
  ; enum		: array [1..maxdim,1..maxdim] of integer



(**********************************************************************)
(*                                                                    *)
(*                 basic routines involving integers                  *)
(*                                                                    *)
(**********************************************************************)

; function sgn ( a : integer ) : sign
  ; begin
      if a>0
      then sgn := 1
      else
      if a<0
      then sgn := -1
      else sgn := 0
    end (* sgn *)

; function modd ( a : integer ) : integer
  (* The smallest positive representative of a mod d. *)
  ; begin
      a := a mod d
    ; if a>0
      then modd := a
      else modd := a+d
    end (* modd *)

; procedure swapints ( var a,b : integer )
  (* Swap two integers. *)
  ; var
      h : integer
  ; begin
      h := a
    ; a := b
    ; b := h
    end (* swapints *)


; function gcd2 ( a,b : integer ) : integer
  (* The g.c.d. of two integers. *)
  ; begin
      a := abs(a)
    ; b := abs(b)
    ; if a<b then swapints(a,b)
    ; while b<>0 do
      begin
        a := a mod b
      ; swapints(a,b)
      end
    ; gcd2 := a
    end (* gcd2 *)


; function gcd3 ( a,b,c : integer ) : integer
  (* The g.c.d. of three integers. *)
  ; begin
      a := abs(a)
    ; b := abs(b)
    ; c := abs(c)
    ; if a<b then swapints(a,b)
    ; if b<c then swapints(b,c)
    ; while c<>0 do
      begin
        a := a mod c
      ; b := b mod c
      ; if a<b
        then swapints(a,c)
        else swapints(b,c)
      end
    ; gcd3 := gcd2(a,b)
    end


(**********************************************************************)
(*                                                                    *)
(*                    making some auxiliary tables                    *)
(*                                                                    *)
(**********************************************************************)

; procedure make2tables
  (* Make a table containing the g.c.d. of every integer with d, *)
  (* and a table containing "inverse integers" of every integer. *)
  ; var
      a : integer
    ; b1,s1,b2,s2,q : integer
  ; begin
      for a := 1 to d do
      begin
        b1 := 0
      ; s1 := d (* b1*a cong. s1 mod d *)
      ; b2 := 1
      ; s2 := a (* b2*a cong. s2 mod d *)
      ; repeat
          q := s1 div s2
        ; b1 := b1-q*b2
        ; s1 := s1-q*s2
        ; swapints(b1,b2)
        ; swapints(s1,s2)
        until s2=0
      ; gcdwithd[a] := s1
      ; invmodd[a] := b1
      end
    ; for a := 1 to d do
      begin
        b1 := invmodd[a] mod (d div gcdwithd[a])
      ; if b1<=0 then b1 := b1+(d div gcdwithd[a])
      ; while gcdwithd[b1]>1 do b1 := b1+(d div gcdwithd[a])
      ; invmodd[a] := b1
      end
    end (* make2tables *)

(**********************************************************************)
(*                                                                    *)
(*           basic procedures involving (signed) 3-vectors            *)
(*                                                                    *)
(**********************************************************************)


; procedure plus ( v : vector3 ; var sv : sgdvector3 )
  (* Provide a 3-vector with a plus sign. *)
  ; begin
      sv.x := v
    ; sv.sg := 1
    end (* plus *)

; function prec ( v1,v2 : vector3 ) : boolean
  (* Test whether the first vector precedes the  *)
  (* second one in the lexicographical ordering. *)
  ; begin
      prec := false
    ; if v1[1]<v2[1]
      then prec := true
      else
        if v1[1]=v2[1] then
        if v1[2]<v2[2]
        then prec := true
        else
          if v1[2]=v2[2] then
          if v1[3]<v2[3]
          then prec := true
    end (* prec *)

; function oppo ( sv1,sv2 : sgdvector3 ) : boolean
  (* Test whether two signed 3-vectors are opposite. *)
  ; begin
      oppo := (sv1.x[1]=sv2.x[1]) and (sv1.x[2]=sv2.x[2])
        and (sv1.x[3]=sv2.x[3]) and (sv1.sg+sv2.sg=0)
    end (* oppo *)

; procedure multscalvec ( sc : integer ; sv : sgdvector3
  ; var sp : sgdvector3 )
  (* Multiply a signed 3-vector by a scalar. *)
  ; var
      i : integer
  ; begin
      for i := 1 to 3 do sp.x[i] := modd(sc*sv.x[i])
    ; sp.sg := sv.sg
    end (* multscalvec *)

; procedure movetox1 ( i : integer ; sv : sgdvector3
  ; var mv : sgdvector3 )
  (* Move x[i] to x[1]. *)
  ; begin
      if i>1 then
      with sv do
      begin
        swapints(x[i],x[1])
      ; sg := -sg
      end
    ; mv := sv
    end (* xitox1 *)

(**********************************************************************)
(*                                                                    *)
(*                 finding the minimal representative                 *)
(*                                                                    *)
(**********************************************************************)

; procedure actd8 ( sv : sgdvector3 ; var min : sgdvector3
                  ; var opp : boolean )
  (* Look at the orbit of sv under D8.  Return a   *)
  (* minimal element in min.  Return in opp        *)
  (* whether the orbit contains opposite elements. *)
  ; begin
      with sv do
      begin
        if modd(-x[2])<x[2] then x[2] := modd(-x[2])
      ; if modd(-x[3])<x[3] then x[3] := modd(-x[3])
      ; if x[3]<x[2] then
        begin
          swapints(x[3],x[2])
        ; sg := -sg
        end
      ; opp := ( x[2]=x[3] )
      end
    ; min := sv
    end (* actd8 *)

; procedure actd8xstab ( sv : sgdvector3 ; var min : sgdvector3
                         ; var opp : boolean )
  (* Look at the orbit of sv under                   *)
  (* D8 x ( stabilizer of x[1] in (Z/dZ)* ).         *)
  (* Return a minimal element in min.  Return in opp *)
  (* whether the orbit contains opposite elements.   *)
  ; var
      sc : integer
    ; sp,mini : sgdvector3
    ; oppi : boolean
  ; begin
      min := sv
    ; opp := false
    ; sc := 1 (* sc will run through integers cong. 1 mod d/sv.x[1] *)
    ; repeat
        if gcdwithd[sc]=1 then (* it's a unit *)
        begin
          multscalvec(sc,sv,sp)
        ; actd8(sp,mini,oppi)
        ; opp := opp or oppi or oppo(mini,min)
        ; if prec(mini.x,min.x) then min := mini
        end
      ; sc := sc+(d div sv.x[1])
      until sc>d
    end (* actd8xstab *)

; procedure acts4xmul ( sv : sgdvector3 ; var min : sgdvector3
                        ; var opp : boolean )
  (* Look at the orbit of sv under S4 x (Z/dZ)*.     *)
  (* Return a minimal element in min.  Return in opp *)
  (* whether the orbit contains opposite elements.   *)
  ; var
      i : integer
    ; sp,mv,mini : sgdvector3
    ; oppi : boolean
  ; begin
      with sv do
      begin
        min := sv
      ; opp := false
      ; for i := 1 to 3 do
        begin
          multscalvec(invmodd[x[i]],sv,sp)
        ; movetox1(i,sp,mv)
        ; actd8xstab(mv,mini,oppi)
        ; opp := opp or oppi or oppo(mini,min)
        ; if prec(mini.x,min.x) then min := mini
        end
      end
    end (* acts4xmul *)

(**********************************************************************)
(*                                                                    *)
(*                     using  the list of orbits                      *)
(*                                                                    *)
(**********************************************************************)

; function orbitnum ( v : vector3 ) : integer
  (* The number in the list of orbits of a minimal representative. *)
  ; var
      lw,md,up : integer
  ; begin
      lw := 1
    ; up := orbits+1
    ; repeat
        md := (lw+up) div 2
      ; if prec(v,orbit[md].x)
        then up := md
        else lw := md
      until up=lw+1
    ; orbitnum := lw
    end (* orbitnum *)

; procedure locate ( v : vector3 ; var on,pr : integer)
  (* Find the orbit number and parameter reference. *)
  ; var
      plusv,min : sgdvector3
    ; opp : boolean
  ; begin
      plus(v,plusv)
    ; acts4xmul(plusv,min,opp)
    ; on := orbitnum(min.x)
    ; pr := min.sg*orbit[on].parref
    end (* parrefofvec *)

(**********************************************************************)
(*                                                                    *)
(*                     making a list of variables                     *)
(*                                                                    *)
(**********************************************************************)

; procedure makeparams
  (* Make a list of parameters, i.e. of orbits whose function      *)
  (* values are not known to vanish or to be equal to the function *)
  (* value of some other orbit.  Set the parref field of orbit.    *)
  ; var
      i : integer
  ; begin
      params := 0
    ; for i := 1 to orbits do
      with orbit[i] do
      if abs(orbref)<i
      then
      begin
        if orbref<>0
        then parref := sgn(orbref)*orbit[abs(orbref)].parref
      end
      else
      begin
        params := params+1
      ; parref := params
      ; param[params] := i
      end
    end (* makeparams *)

(**********************************************************************)
(*                                                                    *)
(*                       input/output routines                        *)
(*                                                                    *)
(**********************************************************************)
 
; procedure readorbits
  ; var i : integer
  ; begin
    ; reset(lnkone,'lnkone') 
    ; readln(lnkone,d)
    ; readln(lnkone,orbits)
    ; for i := 1 to orbits do
      with orbit[i] do
      readln(lnkone,x[1],x[2],x[3],orbref)
    end


(**********************************************************************)
(*                                                                    *)
(*                                                                    *)
(*                                                                    *)
(**********************************************************************)

; function det3x3 ( a : matrix3x3 ) : integer
  (* The determinant of a 3 x 3 matrix. *)
  ; begin
(*    det3x3 := + a[1,1]*a[2,2]*a[3,3] - a[1,1]*a[2,3]*a[3,2]
                - a[1,2]*a[2,1]*a[3,3] + a[1,2]*a[2,3]*a[3,1]
                + a[1,3]*a[2,1]*a[3,2] - a[1,3]*a[2,2]*a[3,1] *)
      det3x3 := + a[1,1]*(a[2,2]*a[3,3]-a[2,3]*a[3,2])
                - a[1,2]*(a[2,1]*a[3,3]-a[2,3]*a[3,1])
                + a[1,3]*(a[2,1]*a[3,2]-a[2,2]*a[3,1])
    end (* det3x3 *)

; procedure mult3x3 ( m1,m2 : matrix3x3 ; var pm : matrix3x3 )
  (* The product of two 3 x 3 matrices. *)
  ; var
      i,j,k : integer
  ; begin
      for i := 1 to 3 do
      for k := 1 to 3 do
      begin
        pm[i,k] := 0
      ; for j := 1 to 3 do pm[i,k] := pm[i,k]+m1[i,j]*m2[j,k]
      end
    end (* mult3x3 *)

; procedure multvec3x3 ( v : vector3 ; m : matrix3x3
  ; var p : vector3 )
  ; var
      i,j : integer
  ; begin
      for j := 1 to 3 do
      begin
        p[j] := 0
      ; for i := 1 to 3 do p[j] := p[j]+v[i]*m[i,j]
      end
    end (* multvec3x3 *)

; procedure swapcols3x3 ( c1,c2 : integer ; var m : matrix3x3 )
  (* Swap two columns of a 3 x 3 matrix. *)
  ; var
      i : integer
  ; begin
      for i := 1 to 3 do swapints(m[i,c1],m[i,c2])
    end (* swapcols3x3 *)

; procedure elimcols3x3 ( r,c1,c2 : integer ; var m : matrix3x3 )
  (* An elimination step on the columns of a 3 x 3 matrix. *)
  ; var
      i,q : integer
  ; begin
      q := m[r,c2] div m[r,c1]
    ; for i := 1 to 3 do m[i,c2] := m[i,c2]-q*m[i,c1]
    end (* elimcols3x3 *)

; procedure swapsim3and3x3 ( i,j : integer ; var v : vector3
    ; var m : matrix3x3 ; var detm : integer )
  (* Simultaneously swap the i-th and j-th coordinate of a
     3-vector, and the i-th and j-th column of a 3 x 3 matrix. *)
  ; var
      k : integer
  ; begin
      swapints(v[i],v[j])
    ; for k := 1 to 3 do swapints(m[k,i],m[k,j])
    ; detm := -detm
    end (* swapsim3and3x3 *)

; procedure redsim3and3x3 ( i,j : integer ; var v : vector3
    ; var m : matrix3x3 )
  (* Reduce the j-th coordinate of a 3-vector modulo
     its i-th coordinate, and apply the opposite
     operation to the columns of a 3 x 3 matrix. *)
  ; var
      k,q : integer
  ; begin
      q := v[j] div v[i]
    ; v[j] := v[j]-q*v[i]
    ; for k := 1 to 3 do m[k,i] := m[k,i]+q*m[k,j]
    end (* redsim3and3x3 *)

; procedure ext3to3x3 ( v : vector3 ; var xm : matrix3x3 )
  (* Find a unimodular matrix whose first column equals v.
     The algorithm is due to F. Beukers.                            *)
  ; var
      i,j,detxm : integer
  ; begin
      for i := 1 to 3 do
      for j := 1 to 3 do
      xm[i,j] := 0
    ; for i := 1 to 3 do
      xm[i,i] := 1
    ; detxm := 1
    ; if abs(v[1])<abs(v[2]) then swapsim3and3x3(1,2,v,xm,detxm)
    ; if abs(v[2])<abs(v[3]) then swapsim3and3x3(2,3,v,xm,detxm)
    ; while v[3]<>0 do
      begin
        for i := 1 to 2 do redsim3and3x3(3,i,v,xm)
      ; if abs(v[1])<abs(v[2])
        then swapsim3and3x3(1,3,v,xm,detxm)
        else swapsim3and3x3(2,3,v,xm,detxm)
      end
    ; if abs(v[1])<abs(v[2]) then swapsim3and3x3(1,2,v,xm,detxm)
    ; while v[2]<>0 do
      begin
        redsim3and3x3(2,1,v,xm)
      ; swapsim3and3x3(1,2,v,xm,detxm)
      end
    ; if v[1]=-1
      then
      begin
        for i := 1 to 3 do xm[i,1] := -xm[i,1]
      ; detxm := -detxm
      end
    ; if detxm=-1
      then for i := 1 to 3 do swapints(xm[i,2],xm[i,3])
    end (* ext3to3x3 *)

(**********************************************************************)
(*                                                                    *)
(*                                                                    *)
(*                                                                    *)
(**********************************************************************)

; function smallrep ( a,m : integer ) : integer
  (* A small representative of a mod m. *)
  ; begin
      m := abs(m)
    ; a := a mod m
    ; if a>m div 2
      then smallrep := a-m
      else 
      if a<-m div 2
      then smallrep := a+m
      else smallrep := a
    end (* smallrep *)

; procedure goodvec ( m : matrix3x3 ; var gv : vector3 )
  (* Find a suitable vector for the reduction of a modular
     symbol to a sum of symbols with smaller determinant. *)
  ; var
      mred : matrix3x3
    ; x : vector3
    ; denom,i : integer
  ; begin
      mred := m
      (* Perform Euclid--Gauss elimination on the first row. *)
    ; if abs(mred[1,1])<abs(mred[1,2]) then swapcols3x3(1,2,mred)
    ; if abs(mred[1,2])<abs(mred[1,3]) then swapcols3x3(2,3,mred)
    ; while mred[1,3]<>0 do
      begin
        elimcols3x3(1,3,1,mred)
      ; elimcols3x3(1,3,2,mred)
      ; if abs(mred[1,1])<abs(mred[1,2])
        then swapcols3x3(1,3,mred)
        else swapcols3x3(2,3,mred)
      end
    ; if abs(mred[1,1])<abs(mred[1,2]) then swapcols3x3(1,2,mred)
    ; while mred[1,2]<>0 do
      begin
        elimcols3x3(1,2,1,mred)
      ; swapcols3x3(1,2,mred)
      end
      (* The Euclid--Gauss elimination on the first row is done. *)
    ; if abs(mred[1,1])>1
      then
      begin
        x[3] := 0
      ; x[2] := 0
      ; x[1] := 1
      ; denom := mred[1,1]
      end
      else
      begin
        (* Perform Euclid--Gauss elimination on the second row. *)
        if abs(mred[2,2])<abs(mred[2,3]) then swapcols3x3(2,3,mred)
      ; while mred[2,3]<>0 do
        begin
          elimcols3x3(2,3,2,mred)
        ; swapcols3x3(2,3,mred)
        end
        (* The Euclid--Gauss elimination on the second row is done. *)
      ; if abs(mred[2,2])>1
        then
        begin
          x[3] := 0
        ; x[2] := 1
        ; x[1] := -mred[1,1]*mred[2,1] (* *x[2] *)
        ; denom := mred[2,2]
        end
        else (* We have |mred[3,3]| = |det(m)| > 1. *)
        begin
          x[3] := 1
        ; x[2] := -mred[2,2]*mred[3,2] (* *x[3] *)
        ; x[1] := -mred[1,1]*(mred[2,1]*x[2]+mred[3,1](* *x[2] *))
        ; denom := mred[3,3]
        end
      end
      (* We have found a vector x such that x.m cong. 0 mod denom. *)
    ; for i := 1 to 3 do x[i] := smallrep(x[i],denom)
    ; multvec3x3(x,m,gv)
    ; for i := 1 to 3 do gv[i] := gv[i] div denom
    end (* goodvec *)

; procedure reprow ( r : integer ; m : matrix3x3 ; v : vector3
                  ; var mv : matrix3x3 )
  (* Replace a row of a 3 x 3 matrix by another vector. *)
  ; var
      i : integer
  ; begin
      mv := m
    ; for i := 1 to 3 do mv[r,i] := v[i]
    end (* reprow *)

; procedure addunimodtorow ( r : integer ; ms : matrix3x3 )
  (* Add a unimodular symbol to a row of the raw Hecke matrix. *)
  ; var
      i,on,pr : integer
    ; v : vector3
  ; begin
      for i := 1 to 3 do v[i] := modd(ms[i,1])
    ; locate(v,on,pr)
    ; if pr<>0 then raw[r,abs(pr)] := raw[r,abs(pr)]+sgn(pr)
    end (* addunimodtorow *)

; procedure addmodsymtorow ( ms : matrix3x3 ; r : integer )
  (* Add a modular symbol to a row of the raw Hecke
     matrix.  In order to do this, reduce the modular
     symbol to a sum of unimodular symbols. *)
  ; var
      i,j,g,detms : integer
    ; mv : matrix3x3
    ; v : vector3
  ; begin
      if det3x3(ms)<>0
      then
      begin
        for i := 1 to 3 do
        begin
          g := gcd3(ms[i,1],ms[i,2],ms[i,3])
        ; if g>1 then for j := 1 to 3 do ms[i,j] := ms[i,j] div g
        end
        (* Now all rows of ms are primitive. *)
      ; detms := det3x3(ms)
      ; if abs(detms)=1
        then
        begin
          if detms=-1
          then for i := 1 to 3 do ms[1,i] := -ms[1,i]
        ; addunimodtorow(r,ms)
        end
        else
        begin
          goodvec(ms,v)
        ; for i := 1 to 3 do
          begin
            reprow(i,ms,v,mv)
          ; addmodsymtorow(mv,r)
          end
        end
      end
    end (* addmodsymtorow *)

; procedure make3x3 ( m11,m12,m13,m21,m22,m23,m31,m32,m33 : integer
                    ; var m : matrix3x3 )
  ; begin
      m[1,1] := m11
    ; m[1,2] := m12
    ; m[1,3] := m13
    ; m[2,1] := m21
    ; m[2,2] := m22
    ; m[2,3] := m23
    ; m[3,1] := m31
    ; m[3,2] := m32
    ; m[3,3] := m33
    end (* thematrix3x3 *)

; procedure calcraw
  (* E_\ell *)
  ; var
      i,s,a,b,c : integer
    ; extmat,cosrep,modsym : matrix3x3
  ; begin
      for s := 1 to dim do
      begin
        for i := 1 to params do raw[s,i] := 0
      ; ext3to3x3(orbit[param[freepar[s]]].x,extmat)
      ; for a := 0 to ell-1 do
        for b := 0 to ell-1 do
        begin
          make3x3(1,0,0,0,1,0,a*d,b,ell,cosrep)
        ; mult3x3(extmat,cosrep,modsym)
        ; addmodsymtorow(modsym,s)
        end
      ; for c := 0 to ell-1 do
        begin
          make3x3(1,0,0,c*d,ell,0,0,0,1,cosrep)
        ; mult3x3(extmat,cosrep,modsym)
        ; addmodsymtorow(modsym,s)
        end
      ; begin
          make3x3(ell,0,0,0,1,0,0,0,1,cosrep)
        ; mult3x3(extmat,cosrep,modsym)
        ; addmodsymtorow(modsym,s)
        end
      end
    end (* calcraw *)

; procedure cook
  ; var
      s,t,i : integer
  ; begin
      for s := 1 to dim do
      for t := 1 to dim do
      begin
        enum[s,t] := 0
      ; for i := 1 to params do
        enum[s,t] := enum[s,t]+raw[s,i]*soln[t,i]
      end
    end

; procedure getg 
  ; var
      s,t : integer
  ; begin
    ; for s := 1 to dim do
      for t := 1 to dim do
      enum[s,t]:=soln[t,freepar[s]]
    end

(**********************************************************************)
(*                                                                    *)
(*                       input/output routines                        *)
(*                                                                    *)
(**********************************************************************)

; procedure readsolnspace
  ; var
    i,s,d2,params2 : integer
  ; begin
    ; reset(lnktwo,'lnktwo')
    ; readln(lnktwo,d2)
    ; readln(lnktwo,params2)
    ; readln(lnktwo,dim)
    ; if dim>0 then
      begin
        for s := 1 to dim do read(lnktwo,freepar[s])
      ; for s := 1 to dim do
        begin
          for i := 1 to params do read(lnktwo,soln[s,i])
        ; readln(lnktwo)
        end
      end
    end

; procedure writemat
  ; var
      s,t : integer
  ; begin
    ; writeln('m=matrix(',dim,',',dim,',X,Y,read());')
    ; for s := 1 to dim do
      for t := 1 to dim do
      begin
        writeln(enum[t,s]:1) (* pari will read column by column *) 
      end
    end

; procedure writeinit
  ; begin
    ; writeln('\prompt=')
    ; writeln('\format=g0.3')
    ; writeln('\p')
    ; writeln('\p')
    ; writeln('\prompt=')
    ; writeln('print()')
    ; writeln('\prompt=d=')
    ; writeln('\prompt=')
    ; writeln(d:1,';')
    ; writeln('\a')
    ; writeln('\prompt=params=')
    ; writeln('\prompt=')
    ; writeln(params:1,':')
    ; writeln('\a')
    ; writeln('\prompt=dim=')
    ; writeln('\prompt=')
    ; writeln(dim:1,';')
    ; writeln('\a')
    ; writemat
    ; writeln('g=m^(-1);') (* The matrix that is needed to correct  
                              the solution matrix and the Hecke matrices. *)
    end

; procedure writeeig
  ; begin
    ; writeln(ell:1,';')
    ; writeln('\w eigen',d:1)
    ; writeln('he=m*g;')
    ; writeln('\r abii',d:1)
    ; writeln('ee=idmat(4);')
    ; writeln(
'if(he==a*ee+b*ii+c*ii*ii+d*ii*ii*ii,a*one+b*rho+c*rho^2+d*rho^3,error);')
    ; writeln('\w eigen',d:1)
    end

; begin
    readorbits
  ; make2tables
  ; makeparams
  ; readsolnspace
  ; getg
  ; writeinit
  ; while not eof do
    begin
      readln(ell)
    ; calcraw
    ; cook
    ; writemat
    ; writeeig
    end
  ; writeln('\q')
  end.
